from functools import partial
from pytz import all_timezones

class Schedule():
    """ The parent of "event" and "recurring" classes."""
    
    def __init__(self, verbose=False):
        self._processes = []
        self._workers = []
        self.verbose = verbose
        
    def timezones(self):
        return all_timezones

    def _print(self, message):
        if self.verbose:
            print(message)

    def run(self):
        try:
            if self._processes:
                [process.start() for process in self._processes]
        except:
            [process.terminate for process in self._processes]
            pass

    def _manifest_function(self, target, args, kwargs):
        try:
            function = partial(target, *args, **kwargs)
            function.__qualname__ = target.__qualname__
            function.__doc__ = target.__doc__
            return function
        except:
            self._processes = []
            raise



from multiprocessing import Process
from time import sleep, time, ctime

from simple_scheduler.base import Schedule

class Recurring(Schedule):
    """ Recurring tasks are those that occur after every "x"-seconds.
        (e.g. script_1 is called every 600 seconds)"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _schedule(self, function, period_in_seconds):
        """
        Parameters
        ----------
        function : a callable function
        period_in_seconds : int
            the time period in seconds

        Returns
        -------
        None.

        """
        while True:
            try:
                s = time() + period_in_seconds
                for x in self._workers:
                    if not x.is_alive():
                        x.join()
                        self._workers.remove(x)
                sleep(s - time())
                p = Process(target = function)
                p.start()
                self._workers.append(p)
                self._print(f"{ctime(time())} :: {function.__qualname__}"+\
                            f" [recurring | {period_in_seconds}-second(s)]")
            except Exception as e:
                self._print(str(e))
                [p.terminate for p in self._workers]
                self._workers = []
                pass

    def add_job(self, target, period_in_seconds, args=(), kwargs={}):
        """
        Assigns an periodic task to a process.

        Parameters
        ----------
        target : a callable function
        period_in_seconds : int
            the time period in seconds to execute this function
        args : tuple(object,), optional
            un-named argumets for the "target" callable
            the default is ()
        kwargs : dict{key:object}, optional
            named argumets for the "target" callable
            the default is {}

        Returns
        -------
        None.

        """
        function = self._manifest_function(target, args, kwargs)
        self._processes.append(Process(target=self._schedule,
                                       name = function.__qualname__,
                                       args=(function, period_in_seconds)))

recurring_scheduler = Recurring(verbose=True)



from pytz import timezone
from datetime import datetime
from multiprocessing import Process
from time import sleep, time, ctime

from simple_scheduler.base import Schedule

class Event(Schedule):
    """ Event occurs at an exact time.
        e.g. scirpt_1 is called at 14:00 and 20:00
        Each event is tried 3-times (but executed only once)."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _schedule(self, function, tz, when):
        """
        Parameters
        ----------
        function : callable function
            name of the function which needs to be scheduled
        tz : str
            timezone
        hour : int
            hour of the time to be scheduled
        minute : int
            minute of the time to be scheduled
        Returns
        -------
        None.
        """
        while True:
            hour = int(datetime.now(timezone(tz)).time().hour)
            minute = int(datetime.now(timezone(tz)).time().minute)
            print(f"{hour}:{minute}", when)
            if f"{hour}:{minute}" in when:
            # if (hour_ == hour) & (minute_ == minute):
                self._print(f"{ctime(time())} :: {function.__qualname__}" +\
                            f" [event @{hour}:{minute} | {tz}]")
                for tries in range(3): # number of attempts for any job
                    try:
                        function()
                        sleep(60) # prevent re-execution of small jobs
                        break
                    except Exception as e:
                        self._print(str(e))
                        sleep(10)
                        continue
            else:
                sleep(55)

    def add_job(self, target, tz, when, args=(), kwargs={}):
        """
        Assigns an event to a process.

        Parameters
        ----------
        target : a callable function
        tz : str
            time zone (call the method .timezones() for more info)
        when : list(str)
            at what precise time(s) should the function be called
            eg. ["12:34","23:45", ...] --> please "only" use 24-hour clock,
                                           with ":" as separator
        args : tuple(object,), optional
            un-named argumets for the "target" callable
            the default is ()
        kwargs : dict{key:object}, optional
            named argumets for the "target" callable
            the default is {}

        Raises
        ------
        Exception
            - If time (in "when"-list) is not a collection of "int:int"
            eg. ["12:30am","2:30 pm", ...] --> please "only" use 24-hour clock,
                                               with ":" as separator

        Returns
        -------
        None.

        """
        function = self._manifest_function(target, args, kwargs)
        try:
            [int(x.split(":")[0]) for x in when]
            [int(x.split(":")[1]) for x in when]
        except ValueError:
            raise Exception('Elements of "when" (list) must be ' +\
                            '["int:int", "int:int", ...]')
        print("step 1")
        print(function, tz, when)
        self._processes.append(Process(target=self._schedule,
                                       name = function.__qualname__,
                                       args=(function, tz, when)))

event_scheduler = Event(verbose=True)


"""
    reference: https://pythonhosted.org/an_example_pypi_project/setuptools.html
"""

import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
print(HERE)
README = (HERE / "README.md").read_text()
# This call to setup() does all the work
setup(
    name="simple_scheduler",
    version="1.0.0",
    description="Easily schedule multiple events and recurring tasks (simultaneously).",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Vernal-Inertia/simple_scheduler",
    author="Vernal Inertia",
    author_email="tanveer2407@gmail.com",
    license="MIT",
    keywords='python, schedule, simple',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha"
    ],
    packages=["simple_scheduler"],
    include_package_data=True,
    install_requires=["pytz"],
    project_urls={
        'Issues': 'https://github.com/Vernal-Inertia/simple_scheduler/issues',
        'Source': 'https://github.com/Vernal-Inertia/simple_scheduler'}
)
